from datetime import datetime, timedelta, timezone
import math
from typing import Sequence, Generator
import time

from pyspark.sql.datasource import DataSourceReader

from pyspark_pi import ds_options, pi, errors, input_partition, context

_ADDITIONAL_PER_PARTITION_DELAY = timedelta(seconds=1)

class PiDataSourceReader(DataSourceReader):
    ctx: context.PiDataSourceContext
    
    def __init__(
            self,
            ctx: context.PiDataSourceContext
    ) -> None:
        self.ctx = ctx

    def partitions(self) -> Sequence[input_partition.PiInputPartition]:
        request_ranges = self._create_request_ranges()
        partitions = self._create_partitions(request_ranges)

        return partitions

    def read(self, partition: input_partition.PiInputPartition) -> Generator[pi.RETURNED_TUPLE_TYPE, None, None]:
        self._wait_to_execute(partition)

        if self.ctx.params.request_type == ds_options.RequestType.RECORDED:
            results = pi.request_recorded_values(self.ctx, partition.request_ranges)
        elif self.ctx.params.request_type == ds_options.RequestType.INTERPOLATED:
            results = pi.request_interpolated_values(self.ctx, partition.request_ranges)
        elif self.ctx.params.request_type == ds_options.RequestType.SUMMARY:
            results = pi.request_summary_values(self.ctx, partition.request_ranges)
        else:
            raise errors.PiDataSourceConfigError(f"Unsupported request type: {self.ctx.params.request_type}")

        yield from results

    def _create_request_ranges(self) -> list[pi.RequestRange]:
        if not self.ctx.points:
            raise errors.PiDataSourceInternalError("No points available to create request ranges.")

        ranges = []
        for point in self.ctx.points:
            freq = point.freq
            total_items = max(1, int((self.ctx.params.end_time - self.ctx.params.start_time) / freq))
            items_per_call = min(total_items, self.ctx.config.max_returned_items_per_call)
            num_ranges = math.ceil(total_items / items_per_call)

            for i in range(num_ranges):
                range_start = self.ctx.params.start_time + i * items_per_call * freq
                range_end = min(range_start + items_per_call * freq, self.ctx.params.end_time)
                ranges.append(pi.RequestRange(point, range_start, range_end))

        return ranges

    def _create_partitions(self, request_ranges: list[pi.RequestRange]) -> list[input_partition.PiInputPartition]:
        partitions = []
        for i in range(0, len(request_ranges), self.ctx.config.rate_limit_max_requests):
            partition_ranges = request_ranges[i:i + self.ctx.config.rate_limit_max_requests]
            partition = input_partition.PiInputPartition(partition_ranges)
            partitions.append(partition)

        # See _wait_to_execute below
        greater_execute_slot_duration = self.ctx.config.rate_limit_duration + _ADDITIONAL_PER_PARTITION_DELAY
        for i, partition in enumerate(partitions):
            partition.execute_offset = i * greater_execute_slot_duration
            partition.execute_cycle_duration = len(partitions) * greater_execute_slot_duration

        return partitions

    def _wait_to_execute(self, partition: input_partition.PiInputPartition) -> None:
        """
        Wait until the current time falls within the allowed execution window for a partition.

        Each partition is assigned a recurring time window (slot) within a fixed-length execution cycle, 
        the duration of which is determined by the total number of partitions. This function ensures that 
        requests for a given partition are spaced apart in time to avoid hitting rate limits when 
        multiple tasks have been dispatched, which all need to send their requests.

        The cycle repeats continuously (e.g., every 5 seconds if there are 5 partitions which each have 
        a 1-second slot), and each partition is assigned a specific offset within that cycle (e.g., 
        partition 2 executes at offset 1s in that 5s cycle). Within that cycle, the partition is allowed 
        to execute only during its designated slot:

            [offset, offset + slot_duration]

        There is also a 1s buffer between slots for additional safety. Nothing is allowed to execute during
        this buffer. So, actually, in the example above, the cycle would be 10 seconds long, with each
        partition executing during a 1s slot, with a 1s buffer after each slot.

        If the current time does not fall within the allowed window, this function will sleep until the
        next cycle where it does.
        """
        if partition.execute_offset is None or partition.execute_cycle_duration is None:
            raise errors.PiDataSourceInternalError("Partition execute_offset and execute_cycle_duration must be set before waiting to execute.")
        
        now = datetime.now(timezone.utc)
        since_epoch = now - datetime(1970, 1, 1, tzinfo=timezone.utc)
        current_time_in_cycle = since_epoch % partition.execute_cycle_duration

        execute_slot_start = partition.execute_offset
        execute_slot_end = execute_slot_start + self.ctx.config.rate_limit_duration

        if not (execute_slot_start <= current_time_in_cycle < execute_slot_end):
            wait_time = (execute_slot_start - current_time_in_cycle).total_seconds()
            if wait_time < 0:
                wait_time += partition.execute_cycle_duration.total_seconds()
            time.sleep(wait_time)


class PiDataSourceError(RuntimeError):
    """Base class for all Pi data source errors."""

class PiDataSourceConfigError(PiDataSourceError):
    """Raised when there is a configuration error for the Pi data source."""

class PiDataSourceHttpError(PiDataSourceError):
    """Raised when there is a connection error to the Pi web API."""
    message: str
    response_message: str
    status_code: int | None

    def __init__(self, message: str, response_message: str, status_code: int | None = None) -> None:
        self.status_code = status_code
        self.message = message
        self.response_message = response_message

        http_hint = f" (HTTP {status_code})" if status_code is not None else ""
        super().__init__(f"{message}{http_hint}: {response_message}")

class PiDataSourceUnexpectedResponseError(PiDataSourceError):
    """Raised when the Pi web API returns a malformed response."""

class PiDataSourceSchemaError(PiDataSourceError):
    """Raised when there is a schema error in the Pi data source."""

class PiDataSourceInternalError(PiDataSourceError):
    """
    Raised when an internal error occurs in the Pi data source.
    These should only be thrown if there is a bug in the library.
    """
    def __init__(self, message: str) -> None:
        super().__init__(f"pyspark_pi internal error, this is almost certainly a bug in the library: {message}.")


import pytest
from datetime import timedelta, datetime, timezone

from pyspark_pi import pi

_NOW = datetime.now()
_TODAY = _NOW.replace(hour=0, minute=0, second=0, microsecond=0)
_YESTERDAY = _TODAY - timedelta(days=1)

@pytest.mark.parametrize("input_str, expected", [
    ("*", _NOW),
    ("", _NOW),
    ("T", _TODAY),
    ("Today", _TODAY),
    ("Y", _YESTERDAY),
    ("Yesterday", _YESTERDAY),
    ("tue", _TODAY - timedelta(days=(_TODAY.weekday() - 1) % 7)),
    ("mon", _TODAY - timedelta(days=(_TODAY.weekday() - 0) % 7)),
    ("wed", _TODAY - timedelta(days=(_TODAY.weekday() - 2) % 7)),
    ("thursday", _TODAY - timedelta(days=(_TODAY.weekday() - 3) % 7)),
    ("jan 1", datetime(_TODAY.year, 1, 1)),
    ("dec 25", datetime(_TODAY.year, 12, 25)),
    ("march 15", datetime(_TODAY.year, 3, 15)),
    ("12/25", datetime(_TODAY.year, 12, 25)),
    ("12-25", datetime(_TODAY.year, 12, 25)),
    ("2020", datetime(2020, _TODAY.month, _TODAY.day)),
    ("1", datetime(_TODAY.year, _TODAY.month, 1)),
    ("15", datetime(_TODAY.year, _TODAY.month, 15)),
    ("2023-08-01T13:45:00Z", datetime(2023, 8, 1, 13, 45, 0, tzinfo=timezone.utc)),
    ("01:00", _TODAY + timedelta(hours=1)),
    ("12:34:56", _TODAY + timedelta(hours=12, minutes=34, seconds=56)),
    ("23:59:59.999", _TODAY + timedelta(hours=23, minutes=59, seconds=59, milliseconds=999)),
    ("T+1h", _TODAY + timedelta(hours=1)),
    ("T-30m", _TODAY - timedelta(minutes=30)),
    ("Yesterday+1h", _YESTERDAY + timedelta(hours=1)),
    ("2020-01-01T00:00:00Z+2h", datetime(2020, 1, 1, 2, 0, 0, tzinfo=timezone.utc)),
    ("12/25+30m", datetime(_TODAY.year, 12, 25) + timedelta(minutes=30)),
    ("12/25-15m", datetime(_TODAY.year, 12, 25) - timedelta(minutes=15)),
    ("1+45m", datetime(_TODAY.year, _TODAY.month, 1) + timedelta(minutes=45)),
    ("15-15m", datetime(_TODAY.year, _TODAY.month, 15) - timedelta(minutes=15)),
    ("2023-08-01T13:45:00+00:00+15m", datetime(2023, 8, 1, 14, 0, 0, tzinfo=timezone.utc)),
    ("2023-08-01T13:45:00-05:00+1h", datetime(2023, 8, 1, 14, 45, 0, tzinfo=timezone(timedelta(hours=-5)))),
    ("mon+1h", (_TODAY - timedelta(days=(_TODAY.weekday() - 0) % 7)) + timedelta(hours=1)),
    ("wed-2h", (_TODAY - timedelta(days=(_TODAY.weekday() - 2) % 7)) - timedelta(hours=2)),
    ("T+1h+30m", _TODAY + timedelta(hours=1, minutes=30)),
    ("T-1h-15m", _TODAY - timedelta(hours=1, minutes=15)),
    ("T+1h-30m", _TODAY + timedelta(minutes=30)),
    ("T-1h+30m", _TODAY - timedelta(minutes=30)),
    ("2023-08-01T00:00:00Z+1h+15m+30s", datetime(2023, 8, 1, 1, 15, 30, tzinfo=timezone.utc)),
    ("2023-08-01T00:00:00Z-1h-15m-30s", datetime(2023, 7, 31, 22, 44, 30, tzinfo=timezone.utc)),
    ("01:00+15m", _TODAY + timedelta(hours=1, minutes=15)),
    ("12:34:56-34m", _TODAY + timedelta(hours=12, minutes=0, seconds=56)),
    ("*+1h", _NOW + timedelta(hours=1)),
    ("*-25m", _NOW - timedelta(minutes=25)),
])
def test_deserialize_pi_time(input_str, expected):
    assert abs(pi.deserialize_pi_time(input_str) - expected) < timedelta(seconds=1)

@pytest.mark.parametrize("invalid_input", [
    "now",
    "tuesdayish",
    "foo",
    "2023-13-01",
    "2023-12-32",
    "2023-08-01T25:00:00Z",
    "monkey",
    "december 32",
    "15:99",
    "9999-99-99",
    "T++1h",
    "T+-",
    "+1h",
    "2023-08-01T13:45:00Z++1h",
    "T+1hour",
    "1.5+1.5",
    "12/25+",
    "12/25-+30m",
    "T+1h+",
    "T-1h-",
])
def test_deserialize_invalid_pi_time(invalid_input):
    with pytest.raises(ValueError):
        pi.deserialize_pi_time(invalid_input)

@pytest.mark.parametrize("input_str, expected", [
    ("", _NOW),
    ("*", _NOW),
    ("T", _TODAY),
    ("Today", _TODAY),
    ("Y", _YESTERDAY),
    ("Yesterday", _YESTERDAY),
    ("tue", _TODAY - timedelta(days=(_TODAY.weekday() - 1) % 7)),
    ("mon", _TODAY - timedelta(days=(_TODAY.weekday() - 0) % 7)),
    ("wed", _TODAY - timedelta(days=(_TODAY.weekday() - 2) % 7)),
    ("thursday", _TODAY - timedelta(days=(_TODAY.weekday() - 3) % 7)),
    ("jan 1", datetime(_TODAY.year, 1, 1)),
    ("dec 25", datetime(_TODAY.year, 12, 25)),
    ("march 15", datetime(_TODAY.year, 3, 15)),
    ("12/25", datetime(_TODAY.year, 12, 25)),
    ("12-25", datetime(_TODAY.year, 12, 25)),
    ("2020", datetime(2020, _TODAY.month, _TODAY.day)),
    ("1", datetime(_TODAY.year, _TODAY.month, 1)),
    ("15", datetime(_TODAY.year, _TODAY.month, 15)),
    ("2023-08-01T13:45:00Z", datetime(2023, 8, 1, 13, 45, 0, tzinfo=timezone.utc)),
    ("01:00", _TODAY + timedelta(hours=1)),
    ("12:34:56", _TODAY + timedelta(hours=12, minutes=34, seconds=56)),
    ("23:59:59.999", _TODAY + timedelta(hours=23, minutes=59, seconds=59, milliseconds=999)),
])
def test_deserialize_timestamp(input_str, expected):
    assert abs(pi.deserialize_pi_timestamp(input_str) - expected) < timedelta(seconds=1)

@pytest.mark.parametrize("invalid_input", [
    "now",
    "todayish",
    "foo",
    "2023-13-01",
    "2023-12-32",
    "2023-08-01T25:00:00Z",
    "monkey",
    "december 32",
    "15:99",
    "9999-99-99",
])
def test_deserialize_invalid_timestamp(invalid_input):
    with pytest.raises(ValueError):
        pi.deserialize_pi_timestamp(invalid_input)


@pytest.mark.parametrize("input_str, expected", [
    ("1:02:03.456", timedelta(hours=1, minutes=2, seconds=3.456)),
    ("2:03", timedelta(hours=2, minutes=3)),
    ("5", timedelta(hours=5)),
    ("1.5h", timedelta(hours=1.5)),
    ("2h30m", timedelta(hours=2, minutes=30)),
    ("+1h-30m", timedelta(minutes=30)),
    ("1h+15m+30s", timedelta(hours=1, minutes=15, seconds=30)),
    ("-1h", timedelta(hours=-1)),
    ("1.25h", timedelta(hours=1.25)),
    ("30m", timedelta(minutes=30)),
    ("45s", timedelta(seconds=45)),
    ("100f", timedelta(milliseconds=100)),
    ("+1h+30m", timedelta(hours=1, minutes=30)),
    ("-2h+90m", timedelta(minutes=-30)),
    ("3h0m", timedelta(hours=3)),
    ("0h+45m", timedelta(minutes=45)),
    ("+1.5h+0.5h", timedelta(hours=2)),
    ("1:2:3", timedelta(hours=1, minutes=2, seconds=3)),
    ("00:00:00", timedelta(0)),
    ("1m30s", timedelta(minutes=1, seconds=30)),
    ("2h15s", timedelta(hours=2, seconds=15)),
    ("0f", timedelta(0)),
])
def test_deserialize_interval(input_str, expected):
    assert abs(pi.deserialize_pi_interval(input_str) - expected) < timedelta(milliseconds=1)

@pytest.mark.parametrize("td, expected", [
    (timedelta(hours=1, minutes=2, seconds=3.456), "1h+2m+3s+456f"),
    (timedelta(hours=2, minutes=30), "2h+30m"),
    (timedelta(minutes=45), "45m"),
    (timedelta(seconds=10), "10s"),
    (timedelta(milliseconds=250), "250f"),
    (timedelta(seconds=0), "0s"),
    (timedelta(hours=-1, minutes=-15), "-1h-15m"),
    (timedelta(minutes=1, seconds=30), "1m+30s"),
    (timedelta(hours=1.5), "1h+30m"),
    (timedelta(milliseconds=1), "1f"),
    (timedelta(seconds=0.001), "1f"),
    (timedelta(seconds=0.999), "999f"),
    (timedelta(seconds=1.999), "1s+999f"),
])
def test_serialize_interval(td, expected):
    assert pi.serialize_pi_interval(td) == expected

@pytest.mark.parametrize("invalid_input", [
    "1y", 
    "2M", 
    "abc", 
    "1x", 
    "1hour", 
    "", 
    ":", 
    "1::", 
    "1:2:3:4",
    "1..5h",
    "1.5.5h",
    "1hh",
    "+-1h",
    "1h30",
    "h",
    "1h+",
    "++1h",
    "--30m",
    "1h 30x",
    "1.5"
])
def test_deserialize_invalid_interval(invalid_input):
    with pytest.raises(ValueError):
        pi.deserialize_pi_interval(invalid_input)



from datetime import timedelta

from .shared import _BaseOptionEnum, _parse_str, _parse_bool, _parse_enum, _parse_positive_int

class AuthMethod(_BaseOptionEnum):
    """
    Enum for the authentication methods supported by the Pi Web API.
    """
    ANONYMOUS = "anonymous"
    BASIC = "basic"
    KERBEROS = "kerberos"
    BEARER = "bearer"

class PiDataSourceConfig:
    host: str
    auth_method: AuthMethod
    username: str | None
    password: str | None
    token: str | None
    verify: bool
    server: str | None
    rate_limit_duration: timedelta
    rate_limit_max_requests: int
    max_returned_items_per_call: int

    def __init__(self, options: dict[str, str]) -> None:
        self.host = _parse_str(options, "host", required=True).rstrip('/')
        self.auth_method = _parse_enum(options, "authMethod", AuthMethod, default="anonymous")
        self.username = _parse_str(options, "username", required=self.auth_method == AuthMethod.BASIC)
        self.password = _parse_str(options, "password", required=self.auth_method == AuthMethod.BASIC)
        self.token = _parse_str(options, "token", required=self.auth_method == AuthMethod.BEARER)
        self.verify = _parse_bool(options, "verify", default="true")
        self.server = _parse_str(options, "server")
        self.rate_limit_duration = timedelta(seconds=_parse_positive_int(options, "rateLimitDuration", default="1")) # In the actual Pi server config, this is specified as a number of seconds, not an AFTimeSpan string.
        self.rate_limit_max_requests = _parse_positive_int(options, "rateLimitMaxRequests", default="1000")
        self.max_returned_items_per_call = _parse_positive_int(options, "maxReturnedItemsPerCall", default="150000")



from enum import Enum
from typing import Any, Type, TypeVar, overload, Literal
from datetime import datetime, timedelta, timezone
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError
import json

from pyspark_pi import pi, errors

class _BaseOptionEnum(Enum):
    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in cls._value2member_map_
    
# TODO: This could be more robust
def parse_paths(options: dict[str, str]) -> list[pi.Path]:
    """
    Parses a string formatted like: ["thing1","thing2","thing3"]
    """
    server = _parse_str(options, "server")
    if "path" in options:
        path = pi.Path(options["path"])
        paths = [path]
    elif "paths" in options:
        paths = json.loads(options["paths"])
        paths = [pi.Path(p) for p in paths]
    else:
        raise errors.PiDataSourceConfigError("A list of point names must be provided using the load() method on the reader.")

    if server:
        for path in paths:
            path.server = server
        return paths
    
    for path in paths:
        if not path.server:
            raise errors.PiDataSourceConfigError("If no server is specified using the 'server' option, all paths must include a server.")

    return paths

@overload
def _parse_str(options: dict[str, str], key: str, *, required: Literal[True], default: str | None = ...) -> str: ...
@overload
def _parse_str(options: dict[str, str], key: str, *, required: bool = ..., default: str) -> str: ...
@overload
def _parse_str(options: dict[str, str], key: str, *, required: bool = ..., default: None = ...) -> str | None: ...

def _parse_str(options: dict[str, str], key: str, *, required: bool = False, default: str | None = None) -> str | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    val = options.get(key, default)
    if val is None:
        return None
    return val.strip()

@overload
def _parse_bool(options: dict[str, str], key: str, *, required: Literal[True], default: str | None = ...) -> bool: ...
@overload
def _parse_bool(options: dict[str, str], key: str, *, required: bool = ..., default: str) -> bool: ...
@overload
def _parse_bool(options: dict[str, str], key: str, *, required: bool = ..., default: None = ...) -> bool | None: ...

def _parse_bool(options: dict[str, str], key: str, *, required: bool = False, default: str | None = "false") -> bool | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    raw = options.get(key, default)
    if raw is None:
        return None
    val = str(raw).strip().lower()
    if val not in {"true", "false"}:
        raise errors.PiDataSourceConfigError(f"Invalid boolean value for {key}: {val}")
    return val == "true"

T = TypeVar("T", bound=_BaseOptionEnum)
@overload
def _parse_enum(options: dict[str, str], key: str, enum_type: Type[T], *, required: Literal[True], default: str | None = ...) -> T: ...
@overload
def _parse_enum(options: dict[str, str], key: str, enum_type: Type[T], *, required: bool = ..., default: str) -> T: ...
@overload
def _parse_enum(options: dict[str, str], key: str, enum_type: Type[T], *, required: bool = ..., default: None = ...) -> T | None: ...

def _parse_enum(options: dict[str, str], key: str, enum_type: Type[T], *, required: bool = False, default: str | None = None) -> T | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    raw = options.get(key, default)
    if raw is None:
        return None
    val = str(raw).strip().lower()
    if not enum_type.has_value(val):
        raise errors.PiDataSourceConfigError(f"Invalid {key} '{val}' specified.")
    return enum_type(val)

@overload
def _parse_datetime(options: dict[str, str], key: str, *, required: Literal[True], default: str | None = ...) -> datetime: ...
@overload
def _parse_datetime(options: dict[str, str], key: str, *, required: bool = ..., default: str) -> datetime: ...
@overload
def _parse_datetime(options: dict[str, str], key: str, *, required: bool = ..., default: None = ...) -> datetime | None: ...

def _parse_datetime(options: dict[str, str], key: str, *, required: bool = False, default: str | None = None) -> datetime | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    raw = options.get(key, default)
    if raw is None:
        return None
    try:
        return pi.deserialize_pi_time(raw)
    except ValueError:
        raise errors.PiDataSourceConfigError(f"Invalid datetime format for {key}: {raw}")

@overload
def _parse_positive_int(options: dict[str, str], key: str, *, required: Literal[True], default: str | None = ...) -> int: ...
@overload
def _parse_positive_int(options: dict[str, str], key: str, *, required: bool = ..., default: str) -> int: ...
@overload
def _parse_positive_int(options: dict[str, str], key: str, *, required: bool = ..., default: None = ...) -> int | None: ...

def _parse_positive_int(options: dict[str, str], key: str, *, required: bool = False, default: str | None = None) -> int | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    val = options.get(key, default)
    if val is None:
        return None
    try:
        val = int(val)
    except ValueError:
        raise errors.PiDataSourceConfigError(f"Invalid integer for {key}: {val}")
    if val <= 0:
        raise errors.PiDataSourceConfigError(f"{key} must be a positive integer.")
    return val

@overload
def _parse_timedelta(options: dict[str, str], key: str, *, required: Literal[True], default: str | None = ...) -> timedelta: ...
@overload
def _parse_timedelta(options: dict[str, str], key: str, *, required: bool = ..., default: str) -> timedelta: ...
@overload
def _parse_timedelta(options: dict[str, str], key: str, *, required: bool = ..., default: None = ...) -> timedelta | None: ...

def _parse_timedelta(options: dict[str, str], key: str, required: bool = False, default: str | None = None) -> timedelta | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    val = options.get(key, default)
    if val is None:
        return None
    try:
        return pi.deserialize_pi_interval(val)
    except ValueError as e:
        raise errors.PiDataSourceConfigError(f"Invalid timedelta format for {key}: {val}. This could be an issue with the des function in the library, not your input. Double-check your input and try formatting it in a different way.") from e

@overload
def _parse_timezone(options: dict[str, str], key: str, *, required: Literal[True], default: str | None = ...) -> ZoneInfo: ...
@overload
def _parse_timezone(options: dict[str, str], key: str, *, required: bool = ..., default: str) -> ZoneInfo: ...
@overload
def _parse_timezone(options: dict[str, str], key: str, *, required: bool = ..., default: None = ...) -> ZoneInfo | None: ...

def _parse_timezone(options: dict[str, str], key: str, required: bool = False, default: str | None = "UTC") -> ZoneInfo | None:
    if required and key not in options:
        raise errors.PiDataSourceConfigError(f"{key} must be specified.")
    val = options.get(key, default)
    if val is None:
        return None
    try:
        return ZoneInfo(val)
    except ZoneInfoNotFoundError:
        raise errors.PiDataSourceConfigError(f"Invalid timezone specified for {key}: {val}")



import re

_PATH_RE = re.compile(
    r'^(?:\\\\(?P<server>[^\\]+)(?=\\))?(?(server)\\|\\?)(?P<rest>[^\\]+(?:\\[^\\]+)*)$'
)

class Path:
    def __init__(self, path: str) -> None:
        m = _PATH_RE.match(path)
        if not m:
            raise ValueError(f"Invalid path format: {path}")
        if m.group("server") == ".":
            raise ValueError(f"Invalid server name: {m.group('server')}. You must use fully-qualified server names.")
        self.server = m.group("server")
        self.parts = m.group("rest").split("\\")

    def __str__(self) -> str:
        parts_str = "\\".join(self.parts)
        return f"\\\\{self.server}\\{parts_str}" if self.server else parts_str


from typing import Any
from datetime import datetime, timedelta

from pyspark.sql.datasource import InputPartition

from pyspark_pi import pi

class PiInputPartition(InputPartition):
    request_ranges: list[pi.RequestRange]
    execute_offset: timedelta | None
    execute_cycle_duration: timedelta | None

    def __init__(self, request_ranges: list[pi.RequestRange]) -> None:
        self.request_ranges = request_ranges
        self.execute_offset = None
        self.execute_cycle_duration = None



"""
https://docs.aveva.com/bundle/pi-web-api-reference/page/help/topics/time-strings.html
"""

from datetime import datetime, timedelta
import dateutil
import re

# ( >︹<)
_TIMESTAMP_ABBREVIATION_RE = re.compile(
    r'''
    ^\s*(
        (?P<star>\*)
        |
        (?P<today>[Tt]oday|[Tt])
        |
        (?P<yesterday>[Yy]esterday|[Yy])
        |
        (?P<weekday>(mon|tue|wed|thu|fri|sat|sun)(day|sday|nesday|rsday|urday)?)
        |
        (?P<monthday>
            (?P<month>(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)(uary|ch|il|e|ust|ember)?)
            \s+
            (?P<md_day>\d{1,2})
        )
        |
        (?P<md>\d{1,2}[-/]\d{1,2})
        |
        (?P<year>\d{4})
        |
        (?P<day>\d{1,2})
    )\s*$
    ''',
    re.IGNORECASE | re.VERBOSE
)

_INTERVAL_SEGMENT_FORMAT_RE = re.compile(
    r'(?P<sign>[+-])?(?P<value>\d+(\.\d+)?)\s*(?P<unit>[yMwdhmsf])(?=[^a-zA-Z]|\Z|\s*)',
    re.IGNORECASE
)

_INTERVAL_COLON_FORMAT_RE = re.compile(
    r'^(?P<h>\d+)(:(?P<m>\d+))?(:(?P<s>\d+(\.\d+)?))?$'
)

_UNIT_TO_SECONDS = {
    'w': 604800,
    'd': 86400,
    'h': 3600,
    'm': 60,
    's': 1,
    'f': 0.001,
}

_UNSUPPORTED_UNITS = {'y', 'M'}

def deserialize_pi_time(value: str) -> datetime:
    """
    Deserialize a full Pi Time string with both a timstamp and an optional interval.
    """
    value = value.strip()
    for i in range(1, len(value) - 1):
        if value[i] in '+-':
            if value[i-1] in "+-" or value[i+1] in "+-":
                raise ValueError(f"Invalid Pi Time format: {value}")
            
            left = value[:i]
            right = value[i+1:]
            try:
                timestamp = deserialize_pi_timestamp(left)
                if value[i] == '+':
                    interval = deserialize_pi_interval(right)
                else:
                    interval = deserialize_pi_interval(f"-{right}")
                return timestamp + interval
            except Exception:
                continue

    return deserialize_pi_timestamp(value)

def deserialize_pi_timestamp(value: str) -> datetime:
    if value == "":
        return datetime.now()
    match = _TIMESTAMP_ABBREVIATION_RE.match(value.strip())
    if not match:
        try:
            return dateutil.parser.parse(value.strip())
        except ValueError:
            raise ValueError(f"Invalid timestamp format: {value}")
    elif match.group("star"):
        return datetime.now()
    elif match.group("today"):
        return datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
    elif match.group("yesterday"):
        return (datetime.now() - timedelta(days=1)).replace(hour=0, minute=0, second=0, microsecond=0)
    elif match.group("weekday"):
        today = datetime.now().replace(hour=0, minute=0, second=0, microsecond=0)
        weekday = match.group("weekday").lower()[:3]
        target_weekday = ["mon", "tue", "wed", "thu", "fri", "sat", "sun"].index(weekday)
        days_ago = (today.weekday() - target_weekday) % 7
        return today - timedelta(days=days_ago)
    elif match.group("monthday"):
        month = match.group("month").lower()[:3]
        target_month = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"].index(month) + 1
        day = int(match.group("md_day"))
        now = datetime.now()
        return datetime(now.year, target_month, day, 0, 0, 0, 0)
    elif match.group("md"):
        month, day = map(int, re.split(r'[-/]', match.group("md")))
        now = datetime.now()
        return datetime(now.year, month, day, 0, 0, 0, 0)
    elif match.group("year"):
        now = datetime.now()
        year = int(match.group("year"))
        return datetime(year, now.month, now.day, 0, 0, 0, 0)
    elif match.group("day"):
        day = int(match.group("day"))
        now = datetime.now()
        return datetime(now.year, now.month, day, 0, 0, 0, 0)
    else:
        raise ValueError(f"Invalid timestamp format: {value}")

def deserialize_pi_interval(value: str) -> timedelta:
    value = value.strip()

    match = _INTERVAL_COLON_FORMAT_RE.match(value)
    if match:
        h = int(match.group("h") or 0)
        m = int(match.group("m") or 0)
        s = float(match.group("s") or 0)
        return timedelta(hours=h, minutes=m, seconds=s)

    matches = list(re.finditer(_INTERVAL_SEGMENT_FORMAT_RE, value))
    if not matches:
        raise ValueError(f"Invalid AFTimeSpan: {value}")
    matches_as_str = "".join(match.group(0) for match in matches)
    if matches_as_str != value:
        raise ValueError(f"Invalid AFTimeSpan: {value}")

    seconds = 0.0
    for part in matches:
        sign = -1 if part.group("sign") == "-" else 1
        num = float(part.group("value"))
        unit = part.group("unit")

        if unit in _UNSUPPORTED_UNITS:
            raise ValueError(f"year and month units are currently not supported: {unit}")
        if unit not in _UNIT_TO_SECONDS:
            raise ValueError(f"Unknown unit: {unit}")

        seconds += sign * num * _UNIT_TO_SECONDS[unit]

    return timedelta(seconds=seconds)

def serialize_pi_interval(td: timedelta) -> str:
    total_seconds = td.total_seconds()
    if total_seconds < 0:
        sign = "-"
        positive = False
    else:
        sign = ""
        positive = True
    total_seconds = abs(total_seconds)

    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    milliseconds = (seconds % 1) * 1000
    seconds = int(seconds)

    parts = []
    if hours:
        parts.append(f"{sign}{int(hours)}h")
        if positive: 
            sign = "+"
    if minutes:
        parts.append(f"{sign}{int(minutes)}m")
        if positive: 
            sign = "+"
    if seconds:
        parts.append(f"{sign}{seconds}s")
        if positive: 
            sign = "+"
    if milliseconds:
        parts.append(f"{sign}{int(milliseconds)}f")

    if not parts:
        parts.append("0s")

    return "".join(parts)


from typing import Any
from datetime import datetime, timezone, timedelta
from zoneinfo import ZoneInfo
import os

import pytest
from pyspark.errors.exceptions.captured import AnalysisException
from pyspark.sql import SparkSession

@pytest.mark.integration
def test_get_recorded(
    spark: SparkSession,
    connection_info: dict[str, str],
    test_data: tuple[type, str, list[dict[str, Any]]],
):
    data_type, point_name, data = test_data

    df = spark.read.format("pi").options(
        **connection_info,
        requestType="recorded",
        startTime=data[0]["Timestamp"],
        endTime=data[-1]["Timestamp"],
        timezone="Etc/UTC"
    ).load(point_name)
    df.show()

    rows = df.collect()

    assert df.count() == len(data)
    assert {row["point"] for row in rows} == {point_name}

    expected_timestamps = [datetime.fromisoformat(row["Timestamp"]).replace(tzinfo=timezone.utc) for row in data]
    # Even if you set spark.sql.session.timeZone to UTC, collect() actually returns timestamps in the local timezone anyway with no tzinfo.
    actual_timestamps = [row["timestamp"].replace(tzinfo=ZoneInfo("America/Chicago")).astimezone(timezone.utc) for row in rows]
    assert actual_timestamps == expected_timestamps

    if data_type is int:
        assert [row["value"] for row in rows] == [d["Value"] for d in data]
    elif data_type is float:
        assert all( abs(row["value"] - d["Value"]) < 0.0001 for row, d in zip(rows, data) )
    elif data_type is str:
        assert [row["value"] for row in rows] == [d["Value"] for d in data]
    elif data_type is datetime:
        # Same tz situation as above
        assert all(
            row["value"].replace(tzinfo=ZoneInfo("America/Chicago")).astimezone(timezone.utc) == datetime.fromisoformat(d["Value"]).replace(tzinfo=timezone.utc)
            for row, d in zip(rows, data)
        )
    elif data_type is bytes:
        assert [row["value"] for row in rows] == [bytes(d["Value"]) for d in data]

@pytest.mark.integration
def test_get_interpolated(
    spark: SparkSession,
    connection_info: dict[str, str],
    test_data: tuple[type, str, list[dict[str, Any]]],
):
    data_type, point_name, data = test_data

    df = spark.read.format("pi").options(
        **connection_info,
        requestType="interpolated",
        interval="30s",
        startTime=data[0]["Timestamp"],
        endTime=data[-1]["Timestamp"],
        timezone="Etc/UTC"
    ).load(point_name)
    df.show()

    rows = df.collect()

    assert df.count() == (len(data) * 2 - 1)
    assert {row["point"] for row in rows} == {point_name}

    for i in range(df.count()):
        actual_timestamp = rows[i]["timestamp"].replace(tzinfo=ZoneInfo("America/Chicago")).astimezone(timezone.utc)
        if i % 2 == 0: # Actual, non-interpolated value
            expected_timestamp = datetime.fromisoformat(data[i // 2]["Timestamp"]).replace(tzinfo=timezone.utc)
            assert actual_timestamp == expected_timestamp
            if data_type is int:
                assert rows[i]["value"] == data[i // 2]["Value"]
            elif data_type is float:
                assert abs(rows[i]["value"] - data[i // 2]["Value"]) < 0.0001
            elif data_type is str:
                assert rows[i]["value"] == data[i // 2]["Value"]
            elif data_type is datetime:
                assert rows[i]["value"].replace(tzinfo=ZoneInfo("America/Chicago")).astimezone(timezone.utc) == datetime.fromisoformat(data[i // 2]["Value"]).replace(tzinfo=timezone.utc)
        else: # Interpolated value
            expected_timestamp = datetime.fromisoformat(data[i // 2]["Timestamp"]).replace(tzinfo=timezone.utc) + timedelta(seconds=30)
            assert actual_timestamp == expected_timestamp
            if data_type is int:
                # Same as previous value
                assert rows[i]["value"] == data[i // 2]["Value"]
            elif data_type is float:
                interpolated_value = (data[i // 2]["Value"] + data[i // 2 + 1]["Value"]) / 2
                assert abs(rows[i]["value"] - interpolated_value) < 0.0001
            elif data_type is str:
                # Same as previous value
                assert rows[i]["value"] == data[i // 2]["Value"]
            elif data_type is datetime:
                # Same as previous value
                assert rows[i]["value"].replace(tzinfo=ZoneInfo("America/Chicago")).astimezone(timezone.utc) == datetime.fromisoformat(data[i // 2]["Value"]).replace(tzinfo=timezone.utc)
        
@pytest.mark.integration
def test_get_average(
    spark: SparkSession,
    connection_info: dict[str, str],
    test_data: tuple[type, str, list[dict[str, Any]]],
):
    data_type, point_name, data = test_data
    if data_type in [str, datetime]:
        return

    df = spark.read.format("pi").options(
        **connection_info,
        requestType="summary",
        summaryType="average",
        calculationBasis="eventweighted",
        summaryDuration="2m",
        startTime=data[0]["Timestamp"],
        endTime=data[-1]["Timestamp"],
        timezone="Etc/UTC"
    ).load(point_name)
    df.show()

    rows = df.collect()

    assert df.count() == (len(data) // 2 - 1)
    assert {row["point"] for row in rows} == {point_name}

    for i in range(df.count()):
        actual_timestamp = rows[i]["timestamp"].replace(tzinfo=ZoneInfo("America/Chicago")).astimezone(timezone.utc)
        expected_timestamp = datetime.fromisoformat(data[i * 2]["Timestamp"]).replace(tzinfo=timezone.utc)
        assert actual_timestamp == expected_timestamp
        
        if data_type is int:
            expected_value = sum(d["Value"] for d in data[i * 2:i * 2 + 2]) // 2
            assert rows[i]["value"] == expected_value
        elif data_type is float:
            expected_value = sum(d["Value"] for d in data[i * 2:i * 2 + 2]) / 2.0
            assert abs(rows[i]["value"] - expected_value) < 0.0001

@pytest.mark.integration
def test_different_point_types(
    spark: SparkSession,
    connection_info: dict[str, str]
):
    points = [os.getenv("PI_WEB_API_INT_POINT_NAME"), os.getenv("PI_WEB_API_FLOAT_POINT_NAME"),]
    if not all(points):
        pytest.skip("Missing env vars for point names for int and float types")

    with pytest.raises(AnalysisException, match="All tags in a given query must be of the same PointType."):
        df = spark.read.format("pi").options(
            **connection_info,
            requestType="recorded",
            startTime="2000-01-01T00:00:00Z",
            endTime="2000-01-01T01:00:00Z",
            timezone="Etc/UTC"
        ).load(points)


import pytest
from pyspark_pi import pi

@pytest.mark.parametrize("path, server, parts", [
    (r"\\Srv\Db\Elem|Attr", "Srv", ["Db", "Elem|Attr"]),
    (r"\\Server123\X\Y\Z", "Server123", ["X", "Y", "Z"]),
    (r"\\{5c64}\Db", "{5c64}", ["Db"]),
    (r"\Db\Elem", None, ["Db", "Elem"]),
    (r"Db\Elem", None, ["Db", "Elem"]),
    (r"Db\Elements[@Category=Tutorial]|Volume", None, ["Db", "Elements[@Category=Tutorial]|Volume"]),
])
def test_path_parse_valid(path, server, parts):
    p = pi.Path(path)
    assert p.server == server
    assert p.parts == parts

@pytest.mark.parametrize("path, expected", [
    (r"\\Srv\Db\Elem|Attr", r"\\Srv\Db\Elem|Attr"),
    (r"\Db\Elem", r"Db\Elem"),
    (r"Db\Elem", r"Db\Elem"),
    (r"\\Server123\X\Y\Z", r"\\Server123\X\Y\Z"),
])
def test_path_str(path, expected):
    assert str(pi.Path(path)) == expected

@pytest.mark.parametrize("invalid_input", [
    r"",
    r"\\",
    r"\\Srv",
    r"\\Srv\\",
    r"Db\\Elem",
    r"\\.\Db\Elem",
    r"\\"
])
def test_path_invalid(invalid_input):
    with pytest.raises(ValueError):
        pi.Path(invalid_input)



from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

from .shared import _BaseOptionEnum, _parse_str, _parse_bool, _parse_enum, _parse_datetime, _parse_positive_int, _parse_timedelta, _parse_timezone

class RequestType(_BaseOptionEnum):
    """
    Enum for the type of request to make to the Pi Web API.
    """
    RECORDED = "recorded"
    INTERPOLATED = "interpolated"
    SUMMARY = "summary"

class BoundaryType(_BaseOptionEnum):
    """
    Defines the behavior of data retrieval at the end points of a specified time range.
    """
    INSIDE = "inside"
    OUTSIDE = "outside"
    INTERPOLATED = "interpolated"

class SummaryType(_BaseOptionEnum):
    """
    Values to indicate which summary type calculation(s) should be performed.
    """
    TOTAL = "total"
    AVERAGE = "average"
    MINIMUM = "minimum"
    MAXIMUM = "maximum"
    RANGE = "range"
    STDDEV = "stddev"
    POPULATION_STDDEV = "populationstddev"
    COUNT = "count"
    PERCENT_GOOD = "percentgood"

class CalculationBasis(_BaseOptionEnum):
    """
    Defines the possible calculation options when performing summary calculations over time-series data.
    """
    TIME_WEIGHTED = "timeweighted"
    EVENT_WEIGHTED = "eventweighted"
    TIME_WEIGHTED_CONTINUOUS = "timeweightedcontinuous"
    TIME_WEIGHTED_DISCRETE = "timeweighteddiscrete"
    EVENT_WEIGHTED_EXCLUDE_MOST_RECENT_EVENT = "eventweightedexcludemostrecentevent"
    EVENT_WEIGHTED_EXCLUDE_EARLIEST_EVENT = "eventweightedexcludeearliestevent"
    EVENT_WEIGHTED_INCLUDE_BOTH_ENDS = "eventweightedincludebothends"

class TimeType(_BaseOptionEnum):
    """
    Defines the timestamp returned for a value when a summary calculation is done.
    """
    AUTO = "auto"
    EARLIEST_TIME = "earliesttime"
    MOST_RECENT_TIME = "mostrecenttime"

class SampleType(_BaseOptionEnum):
    """
    Defines the evaluation of an expression over a time range.
    """
    EXPRESSION_RECORDED_VALUES = "expressionrecordedvalues"
    INTERVAL = "interval"

class PiDataSourceRequestParams:
    request_type: RequestType
    start_time: datetime
    end_time: datetime
    desired_units: str | None
    filter_expression: str | None
    include_filtered_values: bool
    timezone: ZoneInfo | None
    max_count: int
    boundary_type: BoundaryType
    sync_time: str | None
    sync_time_boundary_type: BoundaryType
    interval: timedelta
    summary_type: SummaryType
    calculation_basis: CalculationBasis
    time_type: TimeType
    summary_duration: timedelta | None
    sample_type: SampleType
    sample_interval: timedelta | None

    def __init__(self, options: dict[str, str]) -> None:
        self.request_type = _parse_enum(options, "requestType", RequestType, required=True)
        self.start_time = _parse_datetime(options, "startTime", required=True)
        self.end_time = _parse_datetime(options, "endTime", required=True)
        self.desired_units = _parse_str(options, "desiredUnits")
        self.filter_expression = _parse_str(options, "filterExpression")
        self.include_filtered_values = _parse_bool(options, "includeFilteredValues", default="false")
        self.timezone = _parse_timezone(options, "timezone")
        self.max_count = _parse_positive_int(options, "maxCount", default="150000")
        self.boundary_type = _parse_enum(options, "boundaryType", BoundaryType, default="inside")
        self.sync_time = _parse_str(options, "syncTime")
        self.sync_time_boundary_type = _parse_enum(options, "syncTimeBoundaryType", BoundaryType, default="inside")
        self.interval = _parse_timedelta(options, "interval", default="1h")
        self.summary_type = _parse_enum(options, "summaryType", SummaryType, default="total")
        self.calculation_basis = _parse_enum(options, "calculationBasis", CalculationBasis, default="timeweighted")
        self.time_type = _parse_enum(options, "timeType", TimeType, default="auto")
        self.summary_duration = _parse_timedelta(options, "summaryDuration", required=self.request_type == RequestType.SUMMARY)
        self.sample_type = _parse_enum(options, "sampleType", SampleType, default="expressionrecordedvalues")
        self.sample_interval = _parse_timedelta(options, "sampleInterval", required=self.request_type == RequestType.SUMMARY and self.sample_type == SampleType.INTERVAL)



from .api import RETURNED_TUPLE_TYPE, request_point_metadata, estimate_point_frequencies, request_recorded_values, request_interpolated_values, request_summary_values, RequestRange
from .path import Path
from .point import Point, PointType
from .time import deserialize_pi_time, serialize_pi_interval, deserialize_pi_interval, deserialize_pi_timestamp

__all__ = [
    "RequestRange",
    "request_point_metadata",
    "estimate_point_frequencies",
    "request_recorded_values",
    "request_interpolated_values",
    "request_summary_values",
    "RETURNED_TUPLE_TYPE",
    "Path",
    "Point",
    "PointType",
    "deserialize_pi_time",
    "serialize_pi_interval",
    "deserialize_pi_interval",
    "deserialize_pi_timestamp"
]


from pyspark.sql.datasource import DataSource
from pyspark.sql.types import StructType

from pyspark_pi import ds_options, pi, errors, reader, context, helpers

class PiDataSource(DataSource):
    """
    A Spark data source for the Pi Web API.
    """
    ctx: context.PiDataSourceContext
    schema_inferred: bool

    def __init__(self, options: dict[str, str]) -> None:
        self.ctx = context.PiDataSourceContext(
            config=ds_options.PiDataSourceConfig(options),
            params=ds_options.PiDataSourceRequestParams(options),
            paths=ds_options.parse_paths(options)
        )
        self.schema_inferred = False

    @classmethod
    def name(cls) -> str:
        return "pi"

    def schema(self) -> StructType:
        self.ctx.points, self.ctx.point_type = pi.request_point_metadata(self.ctx)
        self.ctx.points = pi.estimate_point_frequencies(self.ctx)

        self.schema_inferred = True

        return helpers.result_schema(self.ctx.points[0].type.pyspark_type())

    def reader(self, schema: StructType) -> reader.PiDataSourceReader:
        """
        Creates a new PiDataSourceReader with the provided schema.

        The schema() method must always be called whether or not a schema is provided by the user because
        the same call which retrieves the PointTypes from which the schema is derived also retrieves
        the WebIDs of the points to be queried, which are needed. For this reason, if the user does provide a schema,
        schema() is still called here. Thusly, providing a schema really only creates an oportunity for
        an error to be raised if the provided schema does not match that derived from Pi.
        """
        if self.schema_inferred:
            return reader.PiDataSourceReader(self.ctx)

        inferred_schema = self.schema()
        if [field.name for field in schema.fields] != [field.name for field in inferred_schema.fields]:
            raise errors.PiDataSourceSchemaError("Schema mismatch: The provided schema does not match that derived from Pi. Please use the pyspark_pi.result_schema() helper function to generate a compatible schema.")
        if schema["value"].dataType != inferred_schema["value"].dataType:
            raise errors.PiDataSourceSchemaError(f"Schema mismatch: The 'value' field should be of type {inferred_schema['value'].dataType.simpleString()}. This was derived from the PointTypes of the points provided. The provided schema has it as {schema['value'].dataType.simpleString()}.")

        return reader.PiDataSourceReader(self.ctx)


from __future__ import annotations

import requests.auth

from pyspark_pi import ds_options, pi, auth

class PiDataSourceContext:
    config: 'ds_options.PiDataSourceConfig'
    params: 'ds_options.PiDataSourceRequestParams'
    paths: list['pi.Path']
    auth: requests.auth.AuthBase | None
    points: list['pi.Point'] | None
    point_type: 'pi.PointType | None'

    def __init__(self, config: ds_options.PiDataSourceConfig, params: ds_options.PiDataSourceRequestParams, paths: list[pi.Path]) -> None:
        self.config = config
        self.params = params
        self.paths = paths
        self.auth = auth.create_auth(config)
        self.points = None
        self.point_type = None



"""
User-facing helper functions.
"""

from datetime import datetime

from pyspark.sql.types import StructField, StructType, StringType, TimestampType, DataType, IntegerType, FloatType, BooleanType

def result_schema(dtype: type | DataType) -> StructType:
    """
    Returns a result schema based on the provided data type.

    Args:
        dtype (type | DataType): The data type for the value field. Can be int, float, str, datetime, or a pyspark DataType.
    Returns:
        StructType: The result schema.
    Raises:
        ValueError: If the provided data type is not supported.
    """

    if dtype is int:
        out_type = IntegerType()
    elif dtype is float:
        out_type = FloatType()
    elif dtype is str:
        out_type = StringType()
    elif dtype is datetime:
        out_type = TimestampType()
    elif isinstance(dtype, DataType):
        out_type = dtype
    else:
        raise ValueError(f"Unsupported data type: {dtype}")

    return StructType([
        StructField("point", StringType(), nullable=False),
        StructField("timestamp", TimestampType(), nullable=False),
        StructField("value", out_type, nullable=True),
        StructField("units_abbreviation", StringType(), nullable=True),
        StructField("good", BooleanType(), nullable=True),
        StructField("questionable", BooleanType(), nullable=True),
        StructField("substituted", BooleanType(), nullable=True),
        StructField("annotated", BooleanType(), nullable=True),
    ])


from datetime import timedelta, datetime
import time
from typing import Any, TypeAlias

import requests

from pyspark_pi import errors, ds_options, context
from .point import Point, PointType
from .time import serialize_pi_interval

RETURNED_TUPLE_TYPE: TypeAlias = tuple[str, datetime, int | float | str | datetime | bytes | None, str | None, bool, bool, bool, bool]

_SHORT_TIMEOUT = 60
_LONG_TIMEOUT = 60 * 5

_POINT_METADATA_SELECTED_FIELDS = "Items.Object.Name;Items.Object.WebId;Items.Object.PointType;Items.Exception.Errors"
_RECORDED_VALUES_SELECTED_FIELDS = "Items"
_INTERPOLATED_VALUES_SELECTED_FIELDS = _RECORDED_VALUES_SELECTED_FIELDS
_SUMMARY_VALUES_SELECTED_FIELDS = "Items.Value"

class RequestRange:
    def __init__(self, point: Point, start_time: datetime, end_time: datetime):
        self.point = point
        self.start_time = start_time
        self.end_time = end_time

# TODO: There's a limit on the length of a URL which could get hit here
def request_point_metadata(
    ctx: context.PiDataSourceContext
) -> tuple[list[Point], PointType]:
    url = f"{ctx.config.host}/piwebapi/points/multiple?path={'&path='.join([str(path) for path in ctx.paths])}&selectedFields={_POINT_METADATA_SELECTED_FIELDS}"
    res = requests.get(
        url,
        auth=ctx.auth,
        verify=ctx.config.verify,
        timeout=_SHORT_TIMEOUT
    )

    if res.status_code < 200 or res.status_code >= 300:
        message = res.json().get("Message", res.text)
        raise errors.PiDataSourceHttpError("Failed to retrieve point metadata", message, res.status_code)

    res_body = res.json()
    returned_items = res_body.get("Items", [])

    if any("Exception" in item for item in returned_items):
        error_messages = "\n".join(
            item["Exception"]["Errors"][0] for item in returned_items if "Exception" in item
        )
        raise errors.PiDataSourceConfigError("Errors were returned when trying to retrieve point metadata:\n" + error_messages)

    points: list[Point] = []
    for item in returned_items:
        obj = item.get("Object")
        if obj is None:
            raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API when retrieving point metadata: 'Object' field is missing")
        if not all(key in obj for key in ("Name", "WebId", "PointType")):
            raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API when retrieving point metadata: 'Name', 'WebId', or 'PointType' is missing in 'Object'")

        point = Point(
            name=obj.get("Name"),
            web_id=obj.get("WebId"),
            type=PointType(obj.get("PointType"))
        )
        points.append(point)

    if len(set(p.type.pyspark_type() for p in points)) > 1:
        raise errors.PiDataSourceConfigError("All tags in a given query must be of the same PointType")

    return points, points[0].type

def estimate_point_frequencies(
    ctx: context.PiDataSourceContext
) -> list[Point]:
    """
    Estimate the frequency of a point by requesting a count for the last 24 hours.
    This is only necessary for Recorded requests.

    If the point is slower than that, the frequency will be set as once per 24 hours.
    It's better to overestimate the frequency than underestimate it, and a frequency of 24 hours
    for points which are actually slower won't undersize input partitions too much.
    """
    if ctx.points is None or ctx.point_type is None:
        raise errors.PiDataSourceInternalError("Point metadata must be requested before estimating point frequencies")

    if ctx.params.request_type != ds_options.RequestType.RECORDED:
        for point in ctx.points:
            point.freq = ctx.params.interval
        return ctx.points

    batches = [ctx.points[i:i + ctx.config.rate_limit_max_requests] for i in range(0, len(ctx.points), ctx.config.rate_limit_max_requests)]
    for batch in batches:
        url = f"{ctx.config.host}/piwebapi/batch"
        payload: dict[str, dict[str, str]] = {}
        for point in batch:
            payload[point.name] = {
                "Method": "GET",
                "Resource": (
                    f"{ctx.config.host}/piwebapi/streams/{point.web_id}/summary"
                    "?summaryType=Count"
                    "&calculationBasis=EventWeighted"
                    "&startTime=*-24h"
                    "&endTime=*"
                    f"&selectedFields={_SUMMARY_VALUES_SELECTED_FIELDS}"
                )
            }
        headers = {"Content-Type": "application/json", "X-Requested-With": "XMLHttpRequest"}
        res = requests.post(
            url,
            json=payload,
            auth=ctx.auth,
            verify=ctx.config.verify,
            headers=headers,
            timeout=_SHORT_TIMEOUT
        )
        _raise_batch_http_errors(res)

        res_body = res.json()
        results = _parse_batch_summary_results(PointType.INT32, res_body)
        names_counts_map = {name: count for name, _, count, _, _, _, _, _ in results}
        for point in ctx.points:
            count = names_counts_map.get(point.name)
            if count is None or count == 0:
                point.freq = timedelta(days=1)
            else:
                point.freq = timedelta(hours=(24 / count))

        time.sleep(ctx.config.rate_limit_duration.total_seconds())

    return ctx.points

def request_recorded_values(
    ctx: context.PiDataSourceContext,
    request_ranges: list[RequestRange]
) -> list[RETURNED_TUPLE_TYPE]:
    if ctx.point_type is None:
        raise errors.PiDataSourceInternalError("PointType must be known before requesting recorded values")
    
    url = f"{ctx.config.host}/piwebapi/batch"
    headers = {"Content-Type": "application/json", "X-Requested-With": "XMLHttpRequest"}
    payload: dict[str, Any] = {}

    for request_range in request_ranges:
        resource = (
            f"{ctx.config.host}/piwebapi/streams/{request_range.point.web_id}/recorded"
            f"?startTime={request_range.start_time.isoformat().replace('+00:00', 'Z')}"
            f"&endTime={request_range.end_time.isoformat().replace('+00:00', 'Z')}"
            f"&boundaryType={ctx.params.boundary_type.value}"
            f"&selectedFields={_RECORDED_VALUES_SELECTED_FIELDS}"
        )
        if ctx.params.desired_units:
            resource += f"&desiredUnits={ctx.params.desired_units}"
        if ctx.params.filter_expression:
            resource += f"&filterExpression={ctx.params.filter_expression}"
        if ctx.params.include_filtered_values:
            resource += f"&includeFilteredValues={str(ctx.params.include_filtered_values).lower()}"
        if ctx.params.timezone:
            resource += f"&timezone={str(ctx.params.timezone)}"
        if ctx.params.max_count:
            resource += f"&maxCount={ctx.params.max_count}"

        payload[f"{request_range.point.name};{request_range.start_time.isoformat()}"] = {
            "Method": "GET",
            "Resource": resource
        }

    res = requests.post(url, json=payload, auth=ctx.auth, verify=ctx.config.verify, headers=headers, timeout=_LONG_TIMEOUT)
    _raise_batch_http_errors(res)
    return _parse_batch_results(ctx.point_type, res.json())

def request_interpolated_values(
    ctx: context.PiDataSourceContext,
    request_ranges: list[RequestRange]
) -> list[RETURNED_TUPLE_TYPE]:
    if ctx.point_type is None:
        raise errors.PiDataSourceInternalError("PointType must be known before requesting interpolated values")

    url = f"{ctx.config.host}/piwebapi/batch"
    headers = {"Content-Type": "application/json", "X-Requested-With": "XMLHttpRequest"}
    payload: dict[str, Any] = {}

    for request_range in request_ranges:
        resource = (
            f"{ctx.config.host}/piwebapi/streams/{request_range.point.web_id}/interpolated"
            f"?startTime={request_range.start_time.isoformat().replace('+00:00', 'Z')}"
            f"&endTime={request_range.end_time.isoformat().replace('+00:00', 'Z')}"
            f"&interval={serialize_pi_interval(ctx.params.interval)}"
            f"&selectedFields={_INTERPOLATED_VALUES_SELECTED_FIELDS}"
        )
        if ctx.params.sync_time:
            resource += f"&syncTime={ctx.params.sync_time}"
        if ctx.params.sync_time_boundary_type:
            resource += f"&syncTimeBoundaryType={ctx.params.sync_time_boundary_type.value}"
        if ctx.params.desired_units:
            resource += f"&desiredUnits={ctx.params.desired_units}"
        if ctx.params.filter_expression:
            resource += f"&filterExpression={ctx.params.filter_expression}"
        if ctx.params.include_filtered_values:
            resource += f"&includeFilteredValues={str(ctx.params.include_filtered_values).lower()}"
        if ctx.params.timezone:
            resource += f"&timezone={ctx.params.timezone.key}"
        if ctx.params.max_count:
            resource += f"&maxCount={ctx.params.max_count}"

        payload[f"{request_range.point.name};{request_range.start_time.isoformat()}"] = {
            "Method": "GET",
            "Resource": resource
        }

    res = requests.post(url, json=payload, auth=ctx.auth, verify=ctx.config.verify, headers=headers, timeout=_LONG_TIMEOUT)
    _raise_batch_http_errors(res)
    return _parse_batch_results(ctx.point_type, res.json())

def request_summary_values(
    ctx: context.PiDataSourceContext,
    request_ranges: list[RequestRange]
) -> list[RETURNED_TUPLE_TYPE]:
    if ctx.point_type is None:
        raise errors.PiDataSourceInternalError("PointType must be known before requesting summary values")
    if ctx.params.summary_duration is None:
        raise errors.PiDataSourceInternalError("request_summary_values was called but summary_duration is not set in the request parameters")
    
    url = f"{ctx.config.host}/piwebapi/batch"
    headers = {"Content-Type": "application/json", "X-Requested-With": "XMLHttpRequest"}
    payload: dict[str, Any] = {}

    for request_range in request_ranges:
        resource = (
            f"{ctx.config.host}/piwebapi/streams/{request_range.point.web_id}/summary"
            f"?summaryType={ctx.params.summary_type.value}"
            f"&startTime={request_range.start_time.isoformat().replace('+00:00', 'Z')}"
            f"&endTime={request_range.end_time.isoformat().replace('+00:00', 'Z')}"
            f"&calculationBasis={ctx.params.calculation_basis.value}"
            f"&summaryDuration={serialize_pi_interval(ctx.params.summary_duration)}"
            f"&selectedFields={_SUMMARY_VALUES_SELECTED_FIELDS}"
        )
        if ctx.params.time_type:
            resource += f"&timeType={ctx.params.time_type.value}"
        if ctx.params.sample_type:
            resource += f"&sampleType={ctx.params.sample_type.value}"
        if ctx.params.sample_interval:
            resource += f"&sampleInterval={ctx.params.sample_interval}"
        if ctx.params.timezone:
            resource += f"&timezone={str(ctx.params.timezone)}"
        if ctx.params.max_count:
            resource += f"&maxCount={ctx.params.max_count}"
        if ctx.params.filter_expression:
            resource += f"&filterExpression={ctx.params.filter_expression}"

        payload[f"{request_range.point.name};{request_range.start_time.isoformat()}"] = {
            "Method": "GET",
            "Resource": resource
        }

    res = requests.post(url, json=payload, auth=ctx.auth, verify=ctx.config.verify, headers=headers, timeout=_LONG_TIMEOUT)
    _raise_batch_http_errors(res)
    return _parse_batch_summary_results(ctx.point_type, res.json())

def _raise_batch_http_errors(res: requests.Response) -> None:
    if res.status_code < 200 or res.status_code >= 300:
        message = res.json().get("Message", res.text)
        raise errors.PiDataSourceHttpError("Failed to retrieve data from Pi Web API batch endpoint", message, res.status_code)
    res_body = res.json()
    for value in res_body.values():
        if value.get("Status") < 200 or value.get("Status") >= 300:
            content = value.get("Content")
            if not content:
                raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API batch endpoint: 'Content' field is missing in one of the responses")
            message = content.get("Message", str(content))
            raise errors.PiDataSourceHttpError("Failed to retrieve data for one or more points from Pi Web API batch endpoint", message, res.status_code)

def _parse_batch_results(
        point_type: PointType, 
        res_body: dict[str, Any]
) -> list[RETURNED_TUPLE_TYPE]:
    results: list[RETURNED_TUPLE_TYPE] = []
    for key, value_obj in res_body.items():
        point_name = key.split(";")[0]
        content = value_obj.get("Content")
        if not content:
            raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API batch endpoint: 'Content' field is missing in one of the responses")
        items = content.get("Items")
        if not items:
            raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API batch endpoint: 'Items' field is missing in the response content")
        for item in items:
            timestamp = datetime.fromisoformat(item.get("Timestamp"))
            raw_value = item.get("Value")
            if point_type == PointType.DIGITAL:
                value = raw_value.get("Name") if raw_value else None
            elif point_type == PointType.BLOB:
                value = bytes(raw_value) if raw_value else None
            elif point_type == PointType.TIMESTAMP:
                value = datetime.fromisoformat(raw_value) if raw_value else None
            else:
                value = raw_value

            results.append((
                point_name,
                timestamp,
                value,
                item.get("UnitsAbbreviation"),
                bool(item.get("Good")),
                bool(item.get("Questionable")),
                bool(item.get("Substituted")),
                bool(item.get("Annotated")),
            ))
    return results

def _parse_batch_summary_results(
        point_type: PointType,
        res_body: dict[str, Any]
) -> list[RETURNED_TUPLE_TYPE]:
    results: list[RETURNED_TUPLE_TYPE] = []
    for key, value_obj in res_body.items():
        point_name = key.split(";")[0]
        content = value_obj.get("Content")
        if not content:
            raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API batch endpoint: 'Content' field is missing in one of the responses")
        items = content.get("Items")
        if not items:
            raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API batch endpoint: 'Items' field is missing in the response content")
        for item in items:
            inner_value_obj = item.get("Value")
            if not inner_value_obj:
                raise errors.PiDataSourceUnexpectedResponseError("Unexpected response from Pi Web API batch endpoint: 'Value' field is missing in one of the summary items")
            if "Errors" in inner_value_obj:
                error_messages = [err.get("Message", err)[0] for err in inner_value_obj.get("Errors", [])]
                raise errors.PiDataSourceHttpError(f"Error retrieving summary value for point '{point_name}'", "\n".join(error_messages))

            timestamp = datetime.fromisoformat(inner_value_obj.get("Timestamp"))
            raw_value = inner_value_obj.get("Value")
            if point_type == PointType.DIGITAL:
                value = raw_value.get("Name") if raw_value else None
            elif point_type == PointType.BLOB:
                value = bytes(raw_value) if raw_value else None
            elif point_type == PointType.TIMESTAMP:
                value = datetime.fromisoformat(raw_value) if raw_value else None
            else:
                value = raw_value

            results.append((
                point_name,
                timestamp,
                value,
                item.get("UnitsAbbreviation"),
                bool(item.get("Good")),
                bool(item.get("Questionable")),
                bool(item.get("Substituted")),
                bool(item.get("Annotated")),
            ))
    return results




import packaging.version

import pyspark
from pyspark_pi.data_source import PiDataSource
from pyspark_pi.helpers import result_schema

__all__ = ["PiDataSource", "result_schema"]

if packaging.version.parse(pyspark.__version__) < packaging.version.parse("4.0.0"):
    raise ImportError(
        "pyspark-pi requires PySpark 4.0.0 or later. "
    )


import requests.auth

try:
    from requests_kerberos import HTTPKerberosAuth
    _KERBEROS_AVAILABLE = True
except ImportError:
    _KERBEROS_AVAILABLE = False

from pyspark_pi import errors, ds_options

class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token: str) -> None:
        self.token = token

    def __call__(self, req: requests.PreparedRequest) -> requests.PreparedRequest:
        req.headers["authorization"] = "Bearer " + self.token
        return req

def create_auth(
    config: 'ds_options.PiDataSourceConfig'
) -> requests.auth.AuthBase | None:
    if config.auth_method == ds_options.AuthMethod.ANONYMOUS:
        return None
    elif config.auth_method == ds_options.AuthMethod.BASIC:
        if not config.username or not config.password:
            raise errors.PiDataSourceInternalError("create_auth was called with basic auth but username or password is missing.")
        return requests.auth.HTTPBasicAuth(config.username, config.password)
    elif config.auth_method == ds_options.AuthMethod.BEARER:
        if not config.token:
            raise errors.PiDataSourceInternalError("create_auth was called with bearer auth but token is missing.")
        return BearerAuth(config.token)
    elif config.auth_method == ds_options.AuthMethod.KERBEROS:
        if not _KERBEROS_AVAILABLE:
            raise errors.PiDataSourceConfigError("Kerberos authentication is not available. Please install with pip install pyspark-pi[kerberos].")
        return HTTPKerberosAuth()
    else:
        raise errors.PiDataSourceConfigError(f"Unsupported authentication method: {config.auth_method}")


from datetime import timedelta
from enum import Enum

from pyspark.sql.types import DataType, IntegerType, FloatType, StringType, BinaryType, TimestampType

class Point:
    def __init__(self, name: str, web_id: str = None, freq: timedelta = None, type: 'PointType' = None):
        self.name = name
        self.web_id = web_id
        self.freq = freq
        self.type = type

class PointType(Enum):
    INT16 = "Int16"
    INT32 = "Int32"
    FLOAT16 = "Float16"
    FLOAT32 = "Float32"
    FLOAT64 = "Float64"
    STRING = "String"
    DIGITAL = "Digital"
    BLOB = "Blob"
    TIMESTAMP = "Timestamp"

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in cls._value2member_map_
    
    def pyspark_type(self) -> DataType:
        if self == PointType.INT16 or self == PointType.INT32:
            return IntegerType()
        elif self == PointType.FLOAT16 or self == PointType.FLOAT32 or self == PointType.FLOAT64:
            return FloatType()
        elif self == PointType.STRING:
            return StringType()
        elif self == PointType.DIGITAL:
            return StringType()
        elif self == PointType.BLOB:
            return BinaryType()
        elif self == PointType.TIMESTAMP:
            return TimestampType()


import pytest
from datetime import timedelta

from pyspark_pi import reader, ds_options, pi, context

@pytest.fixture
def point():
    return pi.Point(name="p1", freq=timedelta(minutes=1))

@pytest.fixture
def points():
    return [
        pi.Point(name="p1", freq=timedelta(minutes=1), type=pi.PointType.INT32),
        pi.Point(name="p2", freq=timedelta(minutes=1), type=pi.PointType.INT32),
    ]

@pytest.fixture
def default_context():
    ctx = context.PiDataSourceContext(
        ds_options.PiDataSourceConfig({
            "host": "https://example.com",
            "username": "user",
            "password": "pass",
            "verify": "true",
            "server": "server1",
            "rateLimitDuration": "1",
            "rateLimitMaxRequests": 2,
            "maxReturnedItemsPerCall": 10,
        }),
        ds_options.PiDataSourceRequestParams({
            "requestType": "recorded",
            "startTime": "2000-01-01T00:00:00Z",
            "endTime": "2000-01-01T01:00:00Z"
        }),
        ["point"]
    )
    return ctx

def test_single_point_partitioning(point, default_context):
    default_context.points = [point]
    default_context.auth = None
    rdr = reader.PiDataSourceReader(default_context)
    partitions = rdr.partitions()

    expected_ranges = 6
    expected_partitions = 3

    assert len(partitions) == expected_partitions
    assert sum(len(p.request_ranges) for p in partitions) == expected_ranges
    assert partitions[0].request_ranges[0].start_time == default_context.params.start_time
    assert partitions[-1].request_ranges[-1].end_time == default_context.params.end_time
    for request_range in [r for p in partitions for r in p.request_ranges]:
        assert request_range.start_time < request_range.end_time
        assert request_range.end_time - request_range.start_time <= point.freq * default_context.config.max_returned_items_per_call

def test_multiple_points_partitioning(points, default_context):
    default_context.points = points
    default_context.auth = None
    rdr = reader.PiDataSourceReader(default_context)
    partitions = rdr.partitions()

    expected_ranges = 12
    expected_partitions = 6

    assert len(partitions) == expected_partitions
    assert sum(len(p.request_ranges) for p in partitions) == expected_ranges
    assert partitions[0].request_ranges[0].start_time == default_context.params.start_time
    assert partitions[-1].request_ranges[-1].end_time == default_context.params.end_time
    for request_range in [r for p in partitions for r in p.request_ranges]:
        assert request_range.start_time < request_range.end_time
        assert request_range.end_time - request_range.start_time <= request_range.point.freq * default_context.config.max_returned_items_per_call

def test_partition_execute_times_are_staggered(point, default_context):
    default_context.points = [point]
    default_context.auth = None
    rdr = reader.PiDataSourceReader(default_context)
    partitions = rdr.partitions()

    for i, partition in enumerate(partitions):
        assert partition.execute_cycle_duration == len(partitions) * (default_context.config.rate_limit_duration + reader._ADDITIONAL_PER_PARTITION_DELAY)
        previous_partition = partitions[i - 1] if i > 0 else None
        if previous_partition:
            assert partition.execute_offset > previous_partition.execute_offset
            delta = partition.execute_offset - previous_partition.execute_offset
            assert delta == (default_context.config.rate_limit_duration + reader._ADDITIONAL_PER_PARTITION_DELAY)


from .shared import parse_paths
from .config import PiDataSourceConfig, AuthMethod
from .params import PiDataSourceRequestParams, RequestType, BoundaryType, SummaryType, CalculationBasis, SampleType, TimeType

__all__ = [
    "parse_paths",
    "PiDataSourceConfig",
    "AuthMethod",
    "PiDataSourceRequestParams",
    "RequestType",
    "BoundaryType",
    "SummaryType",
    "CalculationBasis",
    "SampleType",
    "TimeType",
]